
module UI.Curses.Internals (gen_decs) where

import Language.Haskell.TH
import Data.Char (isSpace, toLower)
import Data.List (unzip5)
import Text.Regex (mkRegex, matchRegex)

data TType = TVoid
           | TNull
           | TAttr
           | TBool
           | TInt
           | TShort
           | TChType
           | TConstChTypes
           | TConstString
           | TConstWString
           | TWindow
    deriving Eq

gen_decs :: String -> Q [Dec]
gen_decs xs = case dropWhile isSpace xs of
                  "" -> return []
                  xs' -> case break (';' ==) xs' of
                             (func, ';':xs'')
                              | head xs' == '#' -> gen_decs xs''
                              | otherwise -> do this <- do_func func
                                                rest <- gen_decs xs''
                                                return (this ++ rest)
                             _ -> error "Bad dec"

do_func :: String -> Q [Dec]
do_func s
 = case matchRegex (mkRegex "^(.*[ \\*])([a-zA-Z0-9_]*)\\((.*)\\)$") s of
       Just [typ, fun, args] ->
           do_func' (conv_type (reverse $ dropWhile isSpace $ reverse typ),
                     fun,
                     map conv_type (split_args args))
       Just _ -> error "Regexp returned strange result"
       Nothing -> error $ "Bad func: " ++ s

-- XXX Bool return types
do_func' :: (TType, String, [TType]) -> Q [Dec]
do_func' (typ, fun, args) = return [fi, sig, wrap]
    where c_type = foldr arr (AppT (ConT (mkName "IO")) (to_type typ)) (map to_type args)
          c_impspec = "static curses.h " ++ fun
          c_name = mkName ("c_" ++ fun)
          fi = ForeignD (ImportF CCall Safe c_impspec c_name c_type)
          h_type = ctxt_io . ctxt_chtype . ctxt_str . ctxt_chstr
                 $ foldr arr (AppT (VarT (mkName "io")) (to_type typ)) (concat typss)
          ctxt_io = mk_ctxt "io" "MonadIO"
          ctxt_chtype = if TChType `elem` args then mk_ctxt "ch" "IsChType"
                                               else id
          ctxt_str = if TConstString `elem` args then mk_ctxt "str" "IsCString"
                                                 else id
          ctxt_chstr = if TConstChTypes `elem` args
                       then mk_ctxt "chstr" "IsChStr"
                       else id
          h_name = case fun of
                       (c:cs) -> mkName (toLower c:cs)
                       "" -> error "No function name!"
          sig = SigD h_name h_type
          wrap = FunD h_name [cl]
          (pats, typss, before_stmtss, after_stmtss, args_exps)
           = unzip5 $ zipWith handle_arg args [1..]
          cl = Clause (concat pats) (NormalB $ AppE (VarE (mkName "liftIO")) rhs) []
          res_name = mkName "res"
          rhs = DoE (concat before_stmtss
                  ++ [BindS (VarP res_name) call]
                  ++ concat after_stmtss
                  ++ [NoBindS (AppE (VarE (mkName "return")) (VarE res_name))])
          call = foldl AppE (VarE c_name) args_exps

handle_arg :: TType -> Int -> ([Pat], [Type], [Stmt], [Stmt], Exp)
handle_arg t i
 = case t of
       TNull ->
           ([], [], [], [], VarE (mkName "nullPtr"))
       TChType ->
           ([x_pat], [VarT (mkName "ch")], [], [],
            AppE (VarE (mkName "to_ChType")) x_var)
       TConstString ->
           ([x_pat],
            [VarT (mkName "str")],
            [BindS y_pat (AppE (VarE (mkName "to_CString")) x_var)],
            [NoBindS (AppE (AppE (VarE (mkName "used_CString")) x_var) y_var)],
            y_var)
       TConstChTypes ->
           ([x_pat],
            [VarT (mkName "chstr")],
            [BindS y_pat (AppE (VarE (mkName "to_ChStr")) x_var)],
            [NoBindS (AppE (AppE (VarE (mkName "used_ChStr")) x_var) y_var)],
            y_var)
       TBool ->
           ([x_pat],
            [ConT (mkName "Bool")], [], [],
            CondE x_var (VarE (mkName "cTRUE")) (VarE (mkName "cFALSE")))
       _ -> ([x_pat], [to_type t], [], [], x_var)
    where x_pat = VarP (mkName ('x':show i))
          x_var = VarE (mkName ('x':show i))
          y_pat = VarP (mkName ('y':show i))
          y_var = VarE (mkName ('y':show i))

mk_ctxt :: String -> String -> (Type -> Type)
mk_ctxt tv tc = ForallT [mkName tv]
                        [AppT (ConT (mkName tc)) (VarT (mkName tv))]

{-
Make a FI foo called c_foo with type t1 -> t2 -> ... -> tn -> IO tres
Make a type sig fo foo which at its simplest is
    MonadIO io => t1 -> t2 -> ... -> tn -> io tres
Make a function which, in its simplest form, is
    foo a1 a2 ... an = liftIO $ c_foo a1 a2 ... an
-}

arr :: Type -> Type -> Type
arr t1 t2 = AppT (AppT ArrowT t1) t2

split_args :: String -> [String]
split_args "" = []
split_args "void" = [] -- This shouldn't really happen in recursive calls
split_args s = case break (',' ==) s of
                   (x, "") -> [x]
                   (x, _:s') -> x:split_args (dropWhile isSpace s')

conv_type :: String -> TType
conv_type "void" = TVoid
conv_type "void *" = TNull
conv_type "const void *" = TNull
conv_type "attr_t" = TAttr
conv_type "const attr_t" = TAttr
conv_type "bool" = TBool
conv_type "int" = TInt
conv_type "short" = TShort
conv_type "chtype" = TChType
conv_type "const chtype" = TChType
conv_type "const chtype *" = TConstChTypes
conv_type "const char *" = TConstString
conv_type "const wchar_t *" = TConstWString
conv_type "WINDOW *" = TWindow
conv_type "const WINDOW *" = TWindow
-- conv_type "" = T
conv_type s = error $ "conv_type: " ++ s

-- XXX Could use Q monad here
to_type :: TType -> Type
to_type TVoid = TupleT 0
to_type TNull = AppT (ConT (mkName "Ptr")) (TupleT 0)
to_type TAttr = ConT (mkName "Attr")
to_type TBool = ConT (mkName "CBool")
to_type TInt = ConT (mkName "CInt")
to_type TShort = ConT (mkName "CShort")
to_type TChType = ConT (mkName "ChType")
to_type TConstString = ConT (mkName "CString")
to_type TConstWString = AppT (ConT (mkName "Ptr")) (ConT (mkName "CWchar"))
to_type TConstChTypes = AppT (ConT (mkName "Ptr")) (ConT (mkName "ChType"))
to_type TWindow = ConT (mkName "PWindow")

